@@ -40,10 +40,14 @@ import asyncio
 import functools
 import inspect
 import logging
+import math
 logger = logging.getLogger("Edifice")
 import os
+import re
 import typing as tp
 
+import numpy as np
+
 from ._component import BaseComponent, WidgetComponent, RootComponent, register_props
 
 from .qt import QT_VERSION
@@ -54,6 +58,7 @@ if QT_VERSION == "PyQt5":
 else:
     from PySide2 import QtCore, QtWidgets, QtSvg, QtGui
 
+Key = QtCore.Qt.Key
 
 def _ensure_future(fn):
     # Ensures future if fn is a coroutine, otherwise don't modify fn
@@ -186,7 +191,24 @@ class QtWidgetComponent(WidgetComponent):
             Expressed as a dict mapping the name of the context menu entry to either a function
             (which will be called when this entry is clicked) or to another sub context menu.
             For example, {"Copy": copy_fun, "Share": {"Twitter": twitter_share_fun, "Facebook": facebook_share_fun}}
-        on_click: on click callback for the widget. Takes a QMouseEvent object as argument
+        css_class: a string or a list of strings, which will be stored in the "css_class" property of the Qt Widget.
+            This can be used in an application stylesheet, like:
+                QLabel[css_class="heading"] { font-size: 18px; }
+        on_click: callback for click events (mouse pressed and released). Takes a QMouseEvent object as argument
+        on_key_down: callback for key down events (key pressed). Takes a QKeyEvent object as argument.
+            The key() method of QKeyEvent returns the raw key pressed (an element of the QtCore.Qt.Key enum,
+            which is conveniently exposed as edifice.Key).
+            The text() method returns the unicode of the key press, taking modifier keys (e.g. Shift)
+            into account.
+        on_key_up: callback for key up events (key released). Takes a QKeyEvent object as argument.
+        on_mouse_down: callback for mouse down events (mouse pressed). Takes a QMouseEvent object as argument
+        on_mouse_up: callback for mouse up events (mouse released). Takes a QMouseEvent object as argument
+        on_mouse_enter: callback for mouse enter events (triggered once every time mouse enters widget).
+            Takes a QMouseEvent object as argument
+        on_mouse_leave: callback for mouse leave events (triggered once every time mouse leaves widget).
+            Takes a QMouseEvent object as argument
+        on_mouse_move: callback for mouse move events (triggered every time mouse moves within widget).
+            Takes a QMouseEvent object as argument
     """
 
     @register_props
@@ -194,7 +216,16 @@ class QtWidgetComponent(WidgetComponent):
                  tool_tip: tp.Optional[tp.Text] = None,
                  cursor: tp.Optional[tp.Text] = None,
                  context_menu: tp.Optional[ContextMenuType] = None,
-                 on_click: tp.Optional[tp.Callable[[QtGui.QMouseEvent], tp.Any]] = None):
+                 css_class: tp.Optional[tp.Any] = None,
+                 on_click: tp.Optional[tp.Callable[[QtGui.QMouseEvent], tp.Any]] = None,
+                 on_key_down: tp.Optional[tp.Callable[[QtGui.QKeyEvent], tp.Any]] = None,
+                 on_key_up: tp.Optional[tp.Callable[[QtGui.QKeyEvent], tp.Any]] = None,
+                 on_mouse_down: tp.Optional[tp.Callable[[QtGui.QMouseEvent], tp.Any]] = None,
+                 on_mouse_up: tp.Optional[tp.Callable[[QtGui.QMouseEvent], tp.Any]] = None,
+                 on_mouse_enter: tp.Optional[tp.Callable[[QtGui.QMouseEvent], tp.Any]] = None,
+                 on_mouse_leave: tp.Optional[tp.Callable[[QtGui.QMouseEvent], tp.Any]] = None,
+                 on_mouse_move: tp.Optional[tp.Callable[[QtGui.QMouseEvent], tp.Any]] = None,
+                ):
         super().__init__()
         self._height = 0
         self._width = 0
@@ -202,7 +233,21 @@ class QtWidgetComponent(WidgetComponent):
         self._left = 0
         self._size_from_font = None
         self._on_click = None
+        self._on_key_down = None
+        self._default_on_key_down = None
+        self._on_key_up = None
+        self._default_on_key_up = None
+        self._on_mouse_enter = None
+        self._on_mouse_leave = None
+        self._on_mouse_down = None
+        self._on_mouse_up = None
+        self._on_mouse_move = None
         self._widget_children = []
+        self._default_mouse_press_event = None
+        self._default_mouse_release_event = None
+        self._default_mouse_move_event = None
+        self._default_mouse_enter_event = None
+        self._default_mouse_leave_event = None
 
         self._context_menu = None
         self._context_menu_connected = False
@@ -244,10 +289,17 @@ class QtWidgetComponent(WidgetComponent):
             return 0
 
     def _mouse_press(self, ev):
-        pass
+        if self._on_mouse_down is not None:
+            self._on_mouse_down(ev)
+        if self._default_mouse_press_event is not None:
+            self._default_mouse_press_event(ev)
 
     def _mouse_release(self, ev):
         event_pos = ev.pos()
+        if self._on_mouse_up is not None:
+            self._on_mouse_up(ev)
+        if self._default_mouse_release_event is not None:
+            self._default_mouse_release_event(ev)
         geometry = self.underlying.geometry()
 
         if 0 <= event_pos.x() <= geometry.width() and 0 <= event_pos.y() <= geometry.height():
@@ -263,9 +315,80 @@ class QtWidgetComponent(WidgetComponent):
             self._on_click = _ensure_future(on_click)
         else:
             self._on_click = None
+        if self._default_mouse_press_event is None:
+            self._default_mouse_press_event = self.underlying.mousePressEvent
         self.underlying.mousePressEvent = self._mouse_press
+        if self._default_mouse_release_event is None:
+            self._default_mouse_release_event = self.underlying.mouseReleaseEvent
         self.underlying.mouseReleaseEvent = self._mouse_release
 
+    def _set_on_key_down(self, underlying, on_key_down):
+        if self._default_on_key_down is None:
+            self._default_on_key_down = self.underlying.keyPressEvent
+        if on_key_down is not None:
+            self._on_key_down = _ensure_future(on_key_down)
+        else:
+            self._on_key_down = self._default_on_key_down
+        self.underlying.keyPressEvent = self._on_key_down
+
+    def _set_on_key_up(self, underlying, on_key_up):
+        if self._default_on_key_up is None:
+            self._default_on_key_up = self.underlying.keyReleaseEvent
+        if on_key_up is not None:
+            self._on_key_up = _ensure_future(on_key_up)
+        else:
+            self._on_key_up = self._default_on_key_up
+        self.underlying.keyReleaseEvent = self._on_key_up
+
+    def _set_on_mouse_down(self, underlying, on_mouse_down):
+        if on_mouse_down is not None:
+            self._on_mouse_down = _ensure_future(on_mouse_down)
+        else:
+            self._on_mouse_down = None
+        if self._default_mouse_press_event is None:
+            self._default_mouse_press_event = self.underlying.mousePressEvent
+        self.underlying.mousePressEvent = self._mouse_press
+
+    def _set_on_mouse_up(self, underlying, on_mouse_up):
+        if on_mouse_up is not None:
+            self._on_mouse_up = _ensure_future(on_mouse_up)
+        else:
+            self._on_mouse_up = None
+        if self._default_mouse_release_event is None:
+            self._default_mouse_release_event = self.underlying.mouseReleaseEvent
+        self.underlying.mouseReleaseEvent = self._mouse_release
+
+    def _set_on_mouse_enter(self, underlying, on_mouse_enter):
+        if self._default_mouse_enter_event is None:
+            self._default_mouse_enter_event = self.underlying.enterEvent
+        if on_mouse_enter is not None:
+            self._on_mouse_enter = _ensure_future(on_mouse_enter)
+            self.underlying.enterEvent = self._on_mouse_enter
+        else:
+            self._on_mouse_enter = None
+            self.underlying.enterEvent = self._default_mouse_enter_event
+
+    def _set_on_mouse_leave(self, underlying, on_mouse_leave):
+        if self._default_mouse_leave_event is None:
+            self._default_mouse_leave_event = self.underlying.leaveEvent
+        if on_mouse_leave is not None:
+            self._on_mouse_leave = _ensure_future(on_mouse_leave)
+            self.underlying.leaveEvent = self._on_mouse_leave
+        else:
+            self.underlying.leaveEvent = self._default_mouse_leave_event
+            self._on_mouse_leave = None
+
+    def _set_on_mouse_move(self, underlying, on_mouse_move):
+        if self._default_mouse_move_event is None:
+            self._default_mouse_move_event = self.underlying.mouseMoveEvent
+        if on_mouse_move is not None:
+            self._on_mouse_move = _ensure_future(on_mouse_move)
+            self.underlying.mouseMoveEvent = self._on_mouse_move
+            self.underlying.setMouseTracking(True)
+        else:
+            self._on_mouse_move = None
+            self.underlying.mouseMoveEvent = self._default_mouse_move_event
+
     def _gen_styling_commands(self, children, style, underlying, underlying_layout=None):
         commands = []
 
@@ -407,13 +530,33 @@ class QtWidgetComponent(WidgetComponent):
                     style = dict(style)
                 commands.extend(self._gen_styling_commands(children, style, underlying, underlying_layout))
             elif prop == "on_click":
-                if newprops.on_click is not None:
-                    commands.append((self._set_on_click, underlying, newprops.on_click))
-                    if self.props.cursor is not None:
-                        commands.append((underlying.setCursor, QtCore.Qt.PointingHandCursor))
+                commands.append((self._set_on_click, underlying, newprops.on_click))
+                if newprops.on_click is not None and self.props.cursor is not None:
+                    commands.append((underlying.setCursor, QtCore.Qt.PointingHandCursor))
+            elif prop == "on_key_down":
+                commands.append((self._set_on_key_down, underlying, newprops.on_key_down))
+            elif prop == "on_key_up":
+                commands.append((self._set_on_key_up, underlying, newprops.on_key_up))
+            elif prop == "on_mouse_down":
+                commands.append((self._set_on_mouse_down, underlying, newprops.on_mouse_down))
+            elif prop == "on_mouse_up":
+                commands.append((self._set_on_mouse_up, underlying, newprops.on_mouse_up))
+            elif prop == "on_mouse_enter":
+                commands.append((self._set_on_mouse_enter, underlying, newprops.on_mouse_enter))
+            elif prop == "on_mouse_leave":
+                commands.append((self._set_on_mouse_leave, underlying, newprops.on_mouse_leave))
+            elif prop == "on_mouse_move":
+                commands.append((self._set_on_mouse_move, underlying, newprops.on_mouse_move))
             elif prop == "tool_tip":
                 if newprops.tool_tip is not None:
                     commands.append((underlying.setToolTip, newprops.tool_tip))
+            elif prop == "css_class":
+                css_class = newprops.css_class
+                if css_class is None:
+                    css_class = []
+                commands.append((underlying.setProperty, "css_class", css_class))
+                commands.extend([(underlying.style().unpolish, underlying),
+                                 (underlying.style().polish, underlying)])
             elif prop == "cursor":
                 cursor = self.props.cursor or ("default" if self.props.on_click is None else "pointer")
                 commands.append((underlying.setCursor, _CURSORS[cursor]))
@@ -444,13 +587,20 @@ class Window(RootComponent):
 
         if __name__ == "__main__":
             App(MyApp()).start()
+
+    Args:
+        title: the window title
+        icon: the window icon
+        menu: the window's menu bar. In some GUI settings, for example Mac OS,
+            this menu will appear seperately from the window.
+        on_close: event handler for when this window is closed.
     """
 
     @register_props
     def __init__(self, title: tp.Text = "Edifice Application",
-                 on_close: tp.Optional[tp.Callable[[QtGui.QCloseEvent], tp.Any]] = None,
                  icon:tp.Optional[tp.Union[tp.Text, tp.Sequence]] = None,
-                 menu=None):
+                 menu=None,
+                 on_close: tp.Optional[tp.Callable[[QtGui.QCloseEvent], tp.Any]] = None):
         super().__init__()
 
         self._previous_rendering = None
@@ -827,7 +977,8 @@ class TextInput(QtWidgetComponent):
 
     def set_on_change(self, on_change):
         def on_change_fun(text):
-            return _ensure_future(on_change)(text)
+            if text != self.props.text:
+                return _ensure_future(on_change)(text)
         if self._on_change_connected:
             self.underlying.textChanged[str].disconnect()
         self.underlying.textChanged[str].connect(on_change_fun)
@@ -855,6 +1006,8 @@ class TextInput(QtWidgetComponent):
 
         commands = super()._qt_update_commands(children, newprops, newstate, self.underlying)
         commands.append((self.underlying.setText, str(self.props.text)))
+        commands.append((self.underlying.setCursorPosition,
+                         self.underlying.cursorPosition()))
         for prop in newprops:
             if prop == "on_change":
                 commands.append((self.set_on_change, newprops[prop]))
@@ -880,7 +1033,7 @@ class Dropdown(QtWidgetComponent):
     """
 
     @register_props
-    def __init__(self, text: tp.Text = "", selection: tp.Text = "",
+    def __init__(self, selection: tp.Text = "", text: tp.Text = "",
                  options: tp.Optional[tp.Sequence[tp.Text]] = None,
                  editable: bool = False,
                  completer: tp.Optional[Completer] = None,
@@ -1102,6 +1255,17 @@ class Slider(QtWidgetComponent):
         self.underlying = None
         # TODO: let user choose?
         self._granularity = 512
+        if math.isnan(value):
+            raise ValueError("Received nan for value")
+        elif math.isnan(min_value):
+            raise ValueError("Received nan for min_value")
+        elif math.isnan(max_value):
+            raise ValueError("Received nan for max_value")
+        elif min_value == max_value:
+            raise ValueError("min_value must be different from max_value")
+        elif value < min_value or value > max_value:
+            raise ValueError("value must be between min_value and max_value")
+
         if orientation == "horizontal" or orientation == "row":
             self.orientation = QtCore.Qt.Horizontal
         elif orientation == "vertical" or orientation == "column":
@@ -1336,6 +1500,132 @@ class ScrollView(_LinearView):
         return commands
 
 
+def _layout_str_to_grid_spec(layout):
+    """Parses layout to return a grid spec.
+
+    Args:
+        layout: layout string as expected by GridView
+    Returns: (num_rows, num_cols, cell_list), where cell_list is a list of
+        tuples (char_code, row_idx, col_idx, row_span, col_span)
+    """
+    ls = []
+    layout = re.split(";|\n", layout)
+    layout = list(filter(lambda x: x, map(lambda x: x.strip(), layout)))
+    num_rows = len(layout)
+    if num_rows == 0:
+        return (0, 0, [])
+    num_cols = len(layout[0])
+    if num_cols == 0:
+        return (num_rows, 0, [])
+
+    corner = (0, 0)
+    unprocessed = np.ones([num_rows, num_cols])
+    while np.any(unprocessed):
+        char = layout[corner[0]][corner[1]]
+        i = corner[1] + 1
+        for i in range(corner[1] + 1, num_cols + 1):
+            if i >= num_cols or layout[corner[0]][i] != char:
+                break
+        col_span = i - corner[1]
+        j = corner[0] + 1
+        for j in range(corner[0] + 1, num_rows + 1):
+            if j >= num_rows or layout[j][corner[1]] != char:
+                break
+        row_span = j - corner[0]
+
+        ls.append((char, corner[0], corner[1], row_span, col_span))
+        unprocessed[corner[0]: corner[0] + row_span, corner[1]: corner[1] + col_span] = 0
+        corner = np.unravel_index(np.argmax(unprocessed), unprocessed.shape)
+    return (num_rows, num_cols, ls)
+
+
+class GridView(QtWidgetComponent):
+    """Grid layout widget for rendering children on a 2D rectangular grid.
+
+    Grid views allow you to precisely control 2D positioning of widgets.
+    While you can also layout widgets using nested :doc:`Views<edifice.base_components.View>`,
+    specifying the exact location of children relative to each other (with proper alignment)
+    requires extensive fine tuning of style attributes.
+    The GridView allows you to lay out widgets at specified grid indices and size.
+
+    Children will be laid out according to the layout argument.
+    Each child is assigned a character code (by default, the first character of the key;
+    this can be changed via the key_to_code prop).
+    The layout argument describes pictorially where the child should be laid out.
+    For example::
+
+        aabc
+        aabd
+        effg
+
+    describes a layout of 7 children (labeled *a* to *g*) in a 3x4 grid.
+    Child a occupies the top left 2x2 portion of the grid,
+    child b occupies a 2x1 portion of the grid starting from the third column of the first row,
+    etc.
+
+    You can also leave certain spots empty using '_'::
+
+        aa__
+        aabc
+
+    Here is a complete example of using GridView::
+
+        def render(self):
+            return ed.GridView(layout='''
+                789+
+                456-
+                123*
+                00./
+            ''')(
+                ed.Button("7").set_key("7"), ed.Button("8").set_key("8"), ed.Button("9").set_key("9"), ed.Button("+").set_key("+"),
+                ed.Button("4").set_key("4"), ed.Button("5").set_key("5"), ed.Button("6").set_key("6"), ed.Button("-").set_key("-"),
+                ed.Button("1").set_key("1"), ed.Button("2").set_key("2"), ed.Button("3").set_key("3"), ed.Button("*").set_key("*"),
+                ed.Button("0").set_key("0"),                              ed.Button(".").set_key("."), ed.Button("*").set_key("/"),
+            )
+
+    Args:
+
+        layout: description of layout as described above
+        key_to_code: mapping from key to a single character representing that child in the layout string
+    """
+
+    @register_props
+    def __init__(self, layout="", key_to_code=None, **kwargs):
+        super().__init__(**kwargs)
+        self.underlying = None
+        self._previously_rendered = None
+
+    def _initialize(self):
+        self.underlying = QtWidgets.QWidget()
+        self.underlying_layout = QtWidgets.QGridLayout()
+        self.underlying.setObjectName(str(id(self)))
+        self.underlying.setLayout(self.underlying_layout)
+        self.underlying_layout.setContentsMargins(0, 0, 0, 0)
+        self.underlying_layout.setSpacing(0)
+
+    def _clear(self):
+        while self.underlying_layout.takeAt(0) is not None:
+            pass
+
+    def _qt_update_commands(self, children, newprops, newstate):
+        if self.underlying is None:
+            self._initialize()
+        rows, columns, grid_spec = _layout_str_to_grid_spec(self.props.layout)
+        if self.props.key_to_code is None:
+            code_to_child = {c.component._key[0]: c.component for c in children}
+        else:
+            code_to_child = {self.props.key_to_code[c.component._key]: c.component for c in children}
+        grid_spec = [(code_to_child[cell[0]],) + cell[1:] for cell in grid_spec if cell[0] not in " _"]
+        commands = []
+        if grid_spec != self._previously_rendered:
+            commands.append((self._clear,))
+            for child, y, x, dy, dx in grid_spec:
+                commands.append((self.underlying_layout.addWidget, child.underlying, y, x, dy, dx,))
+            self._previously_rendered = grid_spec
+        commands.extend(super()._qt_update_commands(children, newprops, newstate, self.underlying, None))
+        return commands
+
+
 class TabView(_LinearView):
     """Widget with multiple tabs.
 
